/*
 * $Id$
 *
 * Copyright (C) 2001-2003 FhG Fokus
 * Copyright (C) 2004,2005 Free Software Foundation, Inc.
 *
 * This file is part of ser, a free SIP server.
 *
 * ser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * For a license to use the ser software under conditions
 * other than those described here, or to purchase support for this
 * software, please contact iptel.org by e-mail at the following addresses:
 *    info@iptel.org
 *
 * ser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "tls_init.h"
#include "tls_config.h"
#include "../dprint.h"
#include "../mem/shm_mem.h"
#include "../tcp_init.h"
#include "tls_domain.h"
#include <openssl/ssl.h>
#include <netinet/tcp.h>
#include <netinet/ip.h>
#include <unistd.h>

SSL_METHOD     *ssl_methods[TLS_USE_SSLv23 + 1];

/*
 * default context, also for outgoing connections 
 */
SSL_CTX        *default_ctx;


static int
passwd_cb(char *buf, int size, int rwflag, void *filename)
{
    UI             *ui;
    const char     *prompt;
    int             ret;

    ui = UI_new();
    if (ui == NULL)
	goto err;

    prompt = UI_construct_prompt(ui, "passphrase", filename);
    UI_add_input_string(ui, prompt, 0, buf, 0, size - 1);
    ret = UI_process(ui);
    UI_free(ui);
    return strlen(buf);

  err:
    LOG(L_ERR, "tls:pass_cb: Error in pass_cb\n");
    if (ui)
	UI_free(ui);
    return 0;
}


/*
 * Wrappers around SER shared memory functions
 * (which can be macros)
 */

static void    *
ser_malloc(size_t size)
{
    return shm_malloc(size);
}

static void    *
ser_realloc(void *ptr, size_t size)
{
    return shm_realloc(ptr, size);
}

static void
ser_free(void *ptr)
{
    shm_free(ptr);
}


int
tls_init(struct socket_info *si)
{
    DBG("tls_init: Entered\n");

    /*
     * reuse tcp initialization 
     */
    if (tcp_init(si) < 0) {
	LOG(L_ERR, "tls_init: Error while initializing TCP part\n");
	goto error;
    }

    si->proto = PROTO_TLS;
    return 0;

  error:
    if (si->socket != -1) {
	close(si->socket);
	si->socket = -1;
    }
    return -1;
}

/*
 * load a certificate from a file 
 */
static int
load_certificate(SSL_CTX * ctx, char *filename)
{
    DBG("load_certificate: Entered\n");
    if (!SSL_CTX_use_certificate_chain_file(ctx, filename)) {
	LOG(L_ERR,
	    "load_certificate: Unable to load certificate file '%s'\n",
	    filename);
	return -1;
    }

    DBG("load_certificate: '%s' successfuly loaded\n", filename);
    return 0;
}

/*
 * load a private key from a file 
 */
static int
load_private_key(SSL_CTX * ctx, char *filename)
{
    DBG("load_private_key: Entered\n");

    SSL_CTX_set_default_passwd_cb(ctx, passwd_cb);
    SSL_CTX_set_default_passwd_cb_userdata(ctx, filename);

    if (!SSL_CTX_use_PrivateKey_file(ctx, filename, SSL_FILETYPE_PEM)) {
	LOG(L_ERR,
	    "load_private_key: Unable to load private key file '%s'\n",
	    filename);
	return -1;
    }

    if (!SSL_CTX_check_private_key(ctx)) {
	LOG(L_ERR,
	    "load_private_key: Key '%s' does not match the public key of the certificate\n",
	    filename);
	return -1;
    }

    DBG("load_private_key: Key '%s' successfuly loaded\n", filename);
    return 0;
}


static int
load_ca(SSL_CTX * ctx, char *filename)
{
    DBG("load_ca: Entered\n");
    if (!SSL_CTX_load_verify_locations(ctx, filename, 0)) {
	LOG(L_ERR, "load_ca: Unable to load ca '%s'\n", filename);
	return -1;
    }

    DBG("load_ca: CA '%s' successfuly loaded\n", filename);
    return 0;
}


/*
 * initialize ssl methods 
 */
static void
init_ssl_methods(void)
{
    DBG("init_methods: Entered\n");
    ssl_methods[TLS_USE_SSLv2_cli - 1] = SSLv2_client_method();
    ssl_methods[TLS_USE_SSLv2_srv - 1] = SSLv2_server_method();
    ssl_methods[TLS_USE_SSLv2 - 1] = SSLv2_method();

    ssl_methods[TLS_USE_SSLv3_cli - 1] = SSLv3_client_method();
    ssl_methods[TLS_USE_SSLv3_srv - 1] = SSLv3_server_method();
    ssl_methods[TLS_USE_SSLv3 - 1] = SSLv3_method();

    ssl_methods[TLS_USE_TLSv1_cli - 1] = TLSv1_client_method();
    ssl_methods[TLS_USE_TLSv1_srv - 1] = TLSv1_server_method();
    ssl_methods[TLS_USE_TLSv1 - 1] = TLSv1_method();

    ssl_methods[TLS_USE_SSLv23_cli - 1] = SSLv23_client_method();
    ssl_methods[TLS_USE_SSLv23_srv - 1] = SSLv23_server_method();
    ssl_methods[TLS_USE_SSLv23 - 1] = SSLv23_method();
}


/*
 * called once from main.c (main process) 
 */
int
init_tls(void)
{
    struct tls_domain *d;
    DBG("init_tls: Entered\n");

    /*
     * this has to be called before any function calling CRYPTO_malloc,
     * CRYPTO_malloc will set allow_customize in openssl to 0 
     */
    if (!CRYPTO_set_mem_functions(ser_malloc, ser_realloc, ser_free)) {
	LOG(L_ERR,
	    "init_tls: Unable to set the memory allocation functions\n");
	return -1;
    }

    SSL_library_init();
    SSL_load_error_strings();
    init_ssl_methods();

    /*
     * initialize default context first 
     */
    default_ctx = SSL_CTX_new(ssl_methods[tls_method - 1]);
    if (default_ctx == NULL) {
	LOG(L_ERR, "init_tls: Cannot create default ssl context\n");
	return -1;
    }
    if (load_certificate(default_ctx, tls_cert_file) < 0)
	return -1;
    if (tls_ca_file && load_ca(default_ctx, tls_ca_file) < 0)
	return -1;
    if (load_private_key(default_ctx, tls_pkey_file) < 0)
	return -1;

    /*
     * now initialize tls virtual domains 
     */
    d = tls_domains;
    while (d) {
	DBG("init_tls: Processing TLS domain [%s:%d]\n",
	    ip_addr2a(&d->addr), d->port);
	/*
	 * create context 
	 */
	if (d->method == TLS_METHOD_UNSPEC) {
	    DBG("init_tls: No method for tls[%s:%d], using default\n",
		ip_addr2a(&d->addr), d->port);
	    d->method = tls_method;
	}

	d->ctx = SSL_CTX_new(ssl_methods[d->method - 1]);
	if (d->ctx == NULL) {
	    LOG(L_ERR,
		"init_tls: Cannot create ssl context for tls[%s:%d]\n",
		ip_addr2a(&d->addr), d->port);
	    return -1;
	}

	/*
	 * load certificate 
	 */
	if (!d->cert_file) {
	    LOG(L_NOTICE,
		"init_tls: No certificate for tls[%s:%d] defined, using default '%s'\n",
		ip_addr2a(&d->addr), d->port, tls_cert_file);
	    d->cert_file = tls_cert_file;
	}
	if (load_certificate(d->ctx, d->cert_file) < 0)
	    return -1;

	/*
	 * load ca 
	 */
	if (!d->ca_file) {
	    LOG(L_NOTICE,
		"init_tls: No CA for tls[%s:%d] defined, using default '%s'\n",
		ip_addr2a(&d->addr), d->port, tls_ca_file);
	    d->ca_file = tls_ca_file;
	}
	if (d->ca_file && load_ca(d->ctx, d->ca_file) < 0)
	    return -1;
	d = d->next;
    }

    /*
     * load all private keys as the last step (may prompt for password) 
     */
    d = tls_domains;
    while (d) {
	if (!d->pkey_file) {
	    LOG(L_NOTICE,
		"init_tls: No private key for tls[%s:%d] defined, using default '%s'\n",
		ip_addr2a(&d->addr), d->port, tls_pkey_file);
	    d->pkey_file = tls_pkey_file;
	}
	if (load_private_key(d->ctx, d->pkey_file) < 0)
	    return -1;
	d = d->next;
    }
    /*
     * we are all set 
     */
    return 0;
}


/*
 * called from main.c when ser exits (main process) 
 */
void
destroy_tls(void)
{
    struct tls_domain *d;
    DBG("destroy_tls: Entered\n");

    d = tls_domains;
    while (d) {
	if (d->ctx)
	    SSL_CTX_free(d->ctx);
	d = d->next;
    }
    if (default_ctx)
	SSL_CTX_free(default_ctx);
    tls_free_domains();
}
